library("survival")
library("simsurv")
library('ggfortify')
library("ipcwswitch")
require("randomForest")
require('coxed')
require('rpsftm')

##########################################################################
##########################################################################
#
#Simulation Setup
#
##########################################################################
##########################################################################


###########################
#Set up Censore rate and treatment effects
###########################

r_C_list = c(0.25,0.5, 0.75)
delta_list = c(-0.3,-0.5,-0.8)

#One can run loops to go over all r_C&delta combination. For parellel computing purpose I manually set them up
r_C = 0.25
delta = -0.5

#################
#Nuisance
################
n = 400 #sample size
id = 1:n
nRepeat = 1500
HRhat = rep(list(rep(NA,7)),nRepeat)
HRCI = rep(list(matrix(NA,2,7)),nRepeat)

for(iter in 1:nRepeat){
  tryCatch({cat('iter = ',iter)
  
  ################
  #Simulate covariates
  ################
  #Randomization ratio & badprog ratio
  r_R = 0.5; r_B = 0.3
  set.seed(2020+iter)
  trt = rep(0,n)
  trt[sample(x = 1:n,size = n*r_R, replace = F)] = 1
   
  badprog = rep(0,n)                 
  badprog[sample(x = 1:n,size = n*r_B, replace = F)] = 1
  
  #age
  age = rnorm(n, 60, 4)
  
  #biomarker
  beta0 = 0; sigma0 = 1; beta1 = 15; beta2 = 5; beta3 = 0.02
  biomarker = function(t,Badprog,Age){
    return(rnorm(1,beta0, sigma0)+beta1*log(t)+beta2*Badprog+beta3*Age)
  }
  
  ##################
  #Hazard function to simulate event time
  ##################
  lambda = 0.0003; gamma = 0.9; eta = 0.25
  beta0 = rnorm(n,0,1)  
  intercept = rep(1,n)
  covs = data.frame(id,trt,badprog,age,beta0,intercept)
  
  simdat <- simsurv(dist = "weibull", lambdas = lambda, gammas = gamma, betas = c(trt = delta, badprog = 0.5+0.02*beta2, age = 0.02+0.02*beta3, beta0 = 0.02, intercept=0),
                    x = covs[,-1], tde = c(trt = 0,badprog = 0 ,age = 0, beta0 = 0, intercept=0.02*15), tdefunction = "log", interval = c(1e-08, 10*365),maxt = 10*365-1)
  
  d0 =data.frame(simdat,covs[,-c(1,6)])
  
  #Real trt effect
  phi = delta/(gamma+15*0.02)
  
  #Get HR estimation and CI
  m0_cox <- coxph(Surv(eventtime, status)~trt+age+badprog,data =d0)
  summary(m0_cox)  
  
  #####################
  #Switching
  #####################
  #simulate progression & switching probability
  b0 = 5.25
  
  prop_pd = rbeta(n,15,10)
  time_pd = prop_pd*d0$eventtime
  prob_sw = 1/(1+exp(3-0.05*(10+covs[,5]+beta1*log(time_pd)+beta2*covs[,3]+beta3*covs[,4])))
  
  d1 = data.frame(d0,time_pd,prob_sw)
  
  #Simulating event time for switchers
  rBernoulli <- function(p){
    vapply(p, function(x) rbinom(1, 1, x), as.integer(1L))
  }
  
  fac = rmultinom(n,1,c(0.5,0.5))
  f = function(v) return(which(v==1))
  faclevel = apply(fac,2,f)
  fac = t(fac)%*%c(1,0.8)
  
  sw_indicator = rBernoulli(d1$prob_sw)
  sw_indicator[which(trt==1)]=0
  switcher_id = which(sw_indicator == 1)
  
  T_post = d1$eventtime
  T_new =  d1$eventtime
  T_post[switcher_id] = d1$eventtime[switcher_id]-d1$time_pd[switcher_id]
  T_new[switcher_id] = d1$time_pd[switcher_id]+ T_post[switcher_id]*exp(-phi*fac[switcher_id])
  
  d2= data.frame(d1,sw_indicator, T_new, fac )
  
  censor_time = 8*365
  enter_time = runif(n,0,365*2)
  OS = d2$T_new+enter_time
  d2$status = ifelse(OS < censor_time,1,0)
  OS = pmin(OS, censor_time)
  OS = OS-enter_time
  d2$sw_indicator = ifelse((sw_indicator==1)&(OS > time_pd),1,0)
  
  #Random drop_off
  drop_time = OS*rbeta(n,10,10)
  if(round(r_C*n,0)-sum(d2$status==0)>0){
  c_index = sample(which(d2$status==1),size = round(r_C*n,0)-sum(d2$status==0))
  OS[c_index] = drop_time[c_index]
  d2$status[c_index] = 0
  OS = pmin(OS, drop_time)
  d2$sw_indicator = ifelse((sw_indicator==1)&(OS > time_pd),1,0)
  }
  
  d3 = data.frame(d2,OS)
  
  time_sw = d3$time_pd
  time_sw[which((d3$sw_indicator==0)|(d3$trt==1))] = NA
  Simu_data = data.frame(d3[,c(1,3:8,10,13)],time_sw,d3$fac)
  
  ##########################################################################
  ##########################################################################
  #
  #Test existing methods: ITT, censor, IPCW, rpsftm, IPE 
  #
  ##########################################################################
  ##########################################################################
  
  ###################
  #Method 1: ITT
  ###################
  a.data =Simu_data
  
  #Get HR estimation and CI
  itt_cox <- coxph(Surv(OS, status)~trt+age+badprog,data =a.data)
  
  HRhat[[iter]][1] = exp(itt_cox$coefficients[1])
  HRCI[[iter]][,1] = exp(confint(itt_cox))[1,]
  
  #####################
  #Method2: Censor
  #####################
  b.data = Simu_data
  
  #Censor switchers at time_pd
  b.data$OS[b.data$sw_indicator==1] = b.data$time_pd[b.data$sw_indicator==1]
  b.data$status[b.data$sw_indicator==1] = 0
  
  censor_cox <- coxph(Surv(OS, status)~ trt+age+badprog,data = b.data)
  
  HRhat[[iter]][2] = exp(censor_cox$coefficients[1])
  HRCI[[iter]][,2] = exp(confint(censor_cox))[1,]
  
  ####################
  #Method 3: IPCW
  ####################
  #data setup: time interval

  #Visit in every 21 days
  timeN  = c(seq(1,max(Simu_data$OS), by = 21),max(Simu_data$OS)) 
  biomarker = matrix(NA,n,length(timeN))
  w.data = Simu_data[,c(9,1:5,10)]
  for(j in 1:(length(timeN))){
    w.data = data.frame(w.data,rep(timeN[j],n))
    biomarker[,j] = covs[,5]+beta1*log(timeN[j])+beta2*covs[,3]+beta3*covs[,4]
    w.data = data.frame(w.data, biomarker[,j])
  }
  
  
  w.data1 = w.data[,c(1:7,seq(8,dim(w.data)[2],by = 2),seq(9,dim(w.data)[2],by = 2))]
  
  timelist = rep(NA,length(timeN))
  for (j in 1:length(timeN)){timelist[j]=paste('t',as.character(j),sep = '')}
  
  biolist = rep(NA,length(timeN))
  for (j in 1:length(timeN)){biolist[j]=paste('biomarker',as.character(j),sep = '')}
  
  colnames(w.data1) = c(colnames(w.data)[1:7],timelist,biolist)
  
  kept.t <- timesTokeep(w.data1, id = "id",
                        tstart = "t1", tstop = "OS",
                        mes.cov = list(Biomarker = biolist),
                        time.cov = list(timelist))
  
  data.long <- wideToLongTDC(w.data1, id = "id",
                             tstart = "t1", tstop = "OS", event = "status",
                             bas.cov = c("age", "trt", "badprog",'time_sw'),
                             mes.cov = list(Biomarker = biolist),
                             time.cov = list(timelist),
                             times = kept.t[[1]])
  
  tabi <- split(data.long, data.long$id) 
  
  L.tabi <- length(tabi) 
  tablist <- lapply(1:L.tabi, function(i){ 
    refstart <- tabi[[i]]$tstart[1] 
    tabi[[i]]$tstart <- tabi[[i]]$tstart - refstart 
    tabi[[i]]$tstop <- tabi[[i]]$tstop - refstart 
    tabi[[i]]$time_sw <- tabi[[i]]$time_sw - refstart 
    return(tabi[[i]]) 
  })
  data.long <- do.call( rbind, tablist ) # Patients are censored when initiating the other arm treatment, that is, at time swtrtdt data.long2 <- cens.ipw(data.long, id = "id", tstart = "tstart", tstop = "tstop", event = "event", arm = "arm", realtrt = FALSE, censTime ="swtrtdt") 
  
  data.long2 <- cens.ipw(data.long, id = "id", tstart = "tstart", tstop = "tstop", 
                         event = "event", arm = "trt", realtrt = FALSE, 
                         censTime ="time_sw") 
  # 
    rep.times1 <- unique(c(data.long2$tstop[data.long2$cens==1 & data.long2$trt == 1], data.long2$tstop[data.long2$event==1]))
    rep.times2 <- unique(c(data.long2$tstop[data.long2$cens==1 & data.long2$trt == 0], data.long2$tstop[data.long2$event==1]))
  # 
    data.long2$trt = as.factor(data.long2$trt)
    levels(data.long2[,"trt"]) # Now, we can replicate the rows
  # 
    data.rep <- replicRows(data.long2, tstart = 'tstart', tstop = 'tstop',
                           event = 'event', cens = 'cens',
                           times1 = rep.times1, times2 = rep.times2,
                           arm = "trt")
    for(i in 1:dim(data.rep)[1]){
      data.rep$Biomaker[i] = covs[data.rep$id[i],5]+beta1*log(data.rep$tstart[i])+beta2*covs[data.rep$id[i],3]+beta3*covs[data.rep$id[i],4]
    }
  
  data.res <- ipcw(data.long2, id = "id", tstart = tstart, tstop = tstop, cens = cens, 
                   arm = "trt", 
                   bas.cov = c("age", "badprog"), 
                   conf = c("Biomarker"), 
                   trunc = 0.05, type = 'kaplan-meier')
  
  ipcw.cox <- coxph(Surv(tstart, tstop, event) ~ trt +age+badprog+ cluster(id), 
                      data = data.res, weights = data.res$weights) 
 
  HRhat[[iter]][3] = exp(ipcw.cox$coefficients[1])
  HRCI[[iter]][,3] = exp(confint(ipcw.cox))[1,]
  
  ###################
  ##Method 4: rpsftm
  ###################

  c.data =Simu_data
  c.data$rx = rep(0,n)
  c.data$rx[which(c.data$trt == 1)] = 1
  c.data$rx[which((c.data$trt == 0) &(c.data$sw_indicator == 1))] = 1- c.data$time_pd[which((c.data$trt == 0) &(c.data$sw_indicator == 1))]/c.data$OS[which((c.data$trt == 0) &(c.data$sw_indicator == 1))]
  
  rpsftm_fit_lr<-rpsftm(formula=Surv(OS, status)~rand(trt, rx),data=c.data, low_psi = -5,hi_psi=5)
  
  #Get HR estimation and CI
  psi = rpsftm_fit_lr$psi
  c.data0 = c.data
  c.data0$OS = c.data$OS*(1-c.data$rx)+ c.data$OS*c.data$rx*exp(psi)
  c.data0 = c.data0[c.data0$trt==0,]
  
  c.data0 = rbind(c.data0,c.data[which(c.data$trt==1),])
  rpsftm_cox <- coxph(Surv(OS, status)~ trt+age+badprog,data =c.data0)
  
  #summary(rpsftm_cox)  #output provides HR CIs
  #summary(itt_cox)
  #exp(confint(rpsftm_cox))
  
  HRhat[[iter]][4] = exp(rpsftm_cox$coefficients[1])
  HRCI[[iter]][,4] = exp(confint(rpsftm_cox))[1,]
  
  ###################
  ## Method 5: IPE 
  ###################
  est_psi <- function(data, dist){ # data=data set used; dist=distribution used for AFT model
    
    library(survival)
    psi <- NULL
    HR <- NULL
    shape <- NULL
    i <- 0
    diff <- 1
    data$survtime_updated <- data$survtime
    data$status_updated <- data$status
    
    while(diff>1e-05){
      i <- i+1
      fit <- survreg(Surv(survtime_updated, status_updated) ~ arm, data=data, dist=dist)
      psi[i] <- -fit$coefficients[[2]]
      shape[i] <- 1/fit$scale
      HR[i] <- exp(-fit$coefficients[[2]]/fit$scale)
      
      data$survtime_updated <- ifelse(data$xo,
                                      data$xoyrs+(data$survtime-data$xoyrs)*exp(psi[i]), 
                                      data$survtime)
      if(i == 1) diff = 1
      else diff <- abs(exp(psi[i])-exp(psi[i-1]))
    }
    print(psi)  
    list('psi'=psi[length(psi)], 'HR'=HR[length(HR)], "shape"=shape[length(shape)])
  }
  
  
  # Estimate standard error using Bootstrapping
  se_psi_boot <- function(data, dist, B){
    psi_b <- NULL
    for (i in 1:B){
      id <- sample(1:length(data[,1]), length(data[,1]), replace=TRUE)
      data_b <-data[id,]
      psi_b[i] <- est_psi(data_b,dist)[[1]]
    }
    se_psi <- sqrt(sum((psi_b-mean(psi_b))^2)/(B-1))
    list("se_psi"=se_psi)
  }
  
  IPE <- function(data, dist, B){
    psi_est <- est_psi(data=data, dist=dist)$psi
    psi_se <- se_psi_boot(data=data, dist=dist, B=B)$se_psi
    psi_CI <- c(psi_est-1.96*psi_se, psi_est+1.96*psi_se)
    list('psi_est'=psi_est, 'psi_se'=psi_se, 'psi_CI'=psi_CI)
  }
  
  a.data =Simu_data
  arm = a.data$trt
  survtime = a.data$OS
  censyrs = a.data$OS
  status = a.data$status
  xo = a.data$sw_indicator
  xoyrs = a.data$OS-a.data$time_pd
  d.data = data.frame(survtime,status,arm,xo,xoyrs,censyrs)
  
  psi_IPE =est_psi(d.data, 'weibull')[[1]]
  d.data$survtime[Simu_data$sw_indicator==1] = Simu_data$time_pd[Simu_data$sw_indicator==1] + (Simu_data$OS[Simu_data$sw_indicator==1]-Simu_data$time_pd[Simu_data$sw_indicator==1])*exp(psi_IPE[length(psi_IPE)])
  d.data =  d.data[d.data$arm==0,]
  ipe = survfit(Surv(survtime,status)~1,data = d.data)
  
  colnames(d.data) = c("OS","status","trt",NA,NA,NA)
  d.data = rbind(d.data[,1:3],Simu_data[which(Simu_data$trt==1),c(9,2,3)])
  ipe_cox <- coxph(Surv(OS, status)~ trt+age+badprog,data =d.data)

  
  HRhat[[iter]][5] = exp(ipe_cox$coefficients[1])
  HRCI[[iter]][,5] = exp(confint(ipe_cox))[1,]
  
  ##########################
  #Random Forest Prediction
  ##########################
  data = Simu_data
  data0 = Simu_data[which(Simu_data$trt == 0),] 
  data011 = data0[which((data0$sw_indicator == 1)&(data0$status==1)),]
  data001 = data0[which((data0$sw_indicator == 0)&(data0$status==1)),]
  data010 = data0[which((data0$sw_indicator == 1)&(data0$status==0)),]
  data000 = data0[which((data0$sw_indicator == 0)&(data0$status==0)),]
  
  RFfit1 = randomForest(OS ~ age + trt + badprog, data = data001, mtry = 3)
  RFfit0 = randomForest(OS ~ age + trt + badprog, data = data000, mtry = 3)
  U01  = predict(RFfit1, data011)
  U00  = predict(RFfit0, data010)
  data$OS[which(data$trt == 0 & data$sw_indicator == 1 & data$status==1)] = U01
  data$OS[which(data$trt == 0 & data$sw_indicator == 1 & data$status==0)] = U00
  
  rf_cox <- coxph(Surv(OS, status)~ trt + age + badprog,data =data)

  
  HRhat[[iter]][6] = exp(rf_cox$coefficients[1])
  HRCI[[iter]][,6] = exp(confint(rf_cox))[1,]
  
  ###########################
  #Stratified PRSFTM
  ###########################
  c.data = Simu_data
  nlevel = 2
  psi_hat = vector("list",nlevel)
  faclevel =rep(NA,n)
  faclevel[which(c.data$trt == 1)]=0
  faclevel[which((c.data$trt ==0)&(c.data$sw_indicator == 0))]=3
  faclevel[which((c.data$trt ==0)&(c.data$sw_indicator == 1)&(c.data$d3.fac==1))]=1
  faclevel[which((c.data$trt ==0)&(c.data$sw_indicator == 1)&(c.data$d3.fac==0.8))]=2
  
  upper = 0; lower = -2
  c.data$group = as.factor(faclevel)
  
  for(l in 1:nlevel){
    psi_hat[[l]] = seq(lower, upper, 0.02)
  }
  
  d  = expand.grid(psi_hat)
  c.data$U = c.data$OS
  chisq = rep(0,dim(d)[1])
  for(i in 1:dim(d)[1]){
    c.data$U[which((c.data$trt == 1))] = c.data$OS[which(c.data$trt == 1)] *exp(d[i,1])
    for(j in 1: dim(d)[2]){
      c.data$U[which(c.data$group == j)] = c.data$time_sw[which(c.data$group == j)] + (c.data$OS[which(c.data$group == j)] - c.data$time_sw[which(c.data$group == j)]) *exp(d[i,j])
    }
    fit = survdiff(Surv(U,status)~group, data = c.data)
    chisq[i] =fit$chisq
  }
  psi_hat_star = d[which(chisq == min(chisq)),]
  for(j in 1: dim(d)[2]){
    c.data$U[which(c.data$group == j)] = c.data$time_sw[which(c.data$group == j)] + (c.data$OS[which(c.data$group == j)] - c.data$time_sw[which(c.data$group == j)]) *exp(as.numeric(psi_hat_star[j][1,]))
  }
  Srp_cox = coxph(Surv(U, status) ~ trt+age+badprog,data =c.data)
  
  HRhat[[iter]][7] = exp(Srp_cox$coefficients[1])
  HRCI[[iter]][,7] = exp(confint(Srp_cox))[1,]
  },error=function(e){cat("ERROR :",conditionMessage(e), "\n")})
}

write.csv(HRhat,paste('T',as.character(delta),'C',as.character(r_C),"HRhat.csv",sep = '_'))
write.csv(HRCI,paste('T',as.character(delta),'C',as.character(r_C),"HRCI.csv",sep = '_'))


